@@ -70,7 +70,7 @@ typedef struct {
 
 static Bool isdv4Detect(InputInfoPtr);
 static Bool isdv4ParseOptions(InputInfoPtr pInfo);
-static Bool isdv4Init(InputInfoPtr, char* id, float *version);
+static Bool isdv4Init(InputInfoPtr, char* id, size_t id_len, float *version);
 static int isdv4ProbeKeys(InputInfoPtr pInfo);
 static void isdv4InitISDV4(WacomCommonPtr, const char* id, float version);
 static int isdv4GetRanges(InputInfoPtr);
@@ -97,6 +97,7 @@ static int wcmWriteWait(InputInfoPtr pInfo, const char* request);
 		isdv4GetRanges,       /* query ranges */
 		isdv4StartTablet,     /* start tablet */
 		isdv4Parse,
+		NULL,
 	};
 
 static void memdump(InputInfoPtr pInfo, char *buffer, unsigned int len)
@@ -252,7 +253,7 @@ static Bool isdv4ParseOptions(InputInfoPtr pInfo)
  * isdv4Init --
  ****************************************************************************/
 
-static Bool isdv4Init(InputInfoPtr pInfo, char* id, float *version)
+static Bool isdv4Init(InputInfoPtr pInfo, char* id, size_t id_len, float *version)
 {
 	WacomDevicePtr priv = (WacomDevicePtr)pInfo->private;
 	WacomCommonPtr common = priv->common;
@@ -264,8 +265,10 @@ static Bool isdv4Init(InputInfoPtr pInfo, char* id, float *version)
 	if (xf86SetSerialSpeed(pInfo->fd, isdv4data->baudrate) < 0)
 		return !Success;
 
-	if(id)
-		strcpy(id, "ISDV4");
+	if(id) {
+		strncpy(id, "ISDV4", id_len);
+		id[id_len-1] = '\0';
+	}
 	if(version)
 		*version = common->wcmVersion;
 
@@ -613,10 +616,10 @@ static int isdv4ParseTouchPacket(InputInfoPtr pInfo, const unsigned char *data,
 	int channel = 0;
 
 	rc = isdv4ParseTouchData(data, len, common->wcmPktLength, &touchdata);
-	if (rc == -1)
+	if (rc <= 0)
 	{
-		LogMessageVerbSigSafe(X_ERROR, 0, "%s: failed to parse touch data.\n",
-				      pInfo->name);
+		LogMessageVerbSigSafe(X_ERROR, 0, "%s: failed to parse touch data (err %d).\n",
+				      pInfo->name, rc);
 		return -1;
 	}
 
