@@ -251,7 +251,7 @@ static void wcmSendButtonClick(WacomDevicePtr priv, int button, int state)
 static void wcmFingerTapToClick(WacomDevicePtr priv)
 {
 	WacomCommonPtr common = priv->common;
-	WacomDeviceState ds[2] = {{0}}, dsLast[2] = {{0}};
+	WacomDeviceState ds[2] = {}, dsLast[2] = {};
 
 	if (!common->wcmGesture)
 		return;
@@ -311,7 +311,7 @@ static CARD32 wcmSingleFingerTapTimer(OsTimerPtr timer, CARD32 time, pointer arg
 static void wcmSingleFingerTap(WacomDevicePtr priv)
 {
 	WacomCommonPtr common = priv->common;
-	WacomDeviceState ds[2] = {{0}}, dsLast[2] = {{0}};
+	WacomDeviceState ds[2] = {}, dsLast[2] = {};
 
 	getStateHistory(common, ds, ARRAY_SIZE(ds), 0);
 	getStateHistory(common, dsLast, ARRAY_SIZE(dsLast), 1);
@@ -392,7 +392,7 @@ void wcmCancelGesture(InputInfoPtr pInfo)
 void wcmGestureFilter(WacomDevicePtr priv, int touch_id)
 {
 	WacomCommonPtr common = priv->common;
-	WacomDeviceState ds[2] = {{0}}, dsLast[2] = {{0}};
+	WacomDeviceState ds[2] = {}, dsLast[2] = {};
 
 	getStateHistory(common, ds, ARRAY_SIZE(ds), 0);
 	getStateHistory(common, dsLast, ARRAY_SIZE(dsLast), 1);
@@ -584,7 +584,7 @@ static void wcmSendScrollEvent(WacomDevicePtr priv, int dist,
 	WacomCommonPtr common = priv->common;
 	int count = (int)((1.0 * abs(dist)/
 		common->wcmGestureParameters.wcmScrollDistance) + 0.5);
-	WacomDeviceState ds[2] = {{0}};
+	WacomDeviceState ds[2] = {};
 
 	getStateHistory(common, ds, ARRAY_SIZE(ds), 0);
 
@@ -610,12 +610,13 @@ static void wcmSendScrollEvent(WacomDevicePtr priv, int dist,
 static void wcmFingerScroll(WacomDevicePtr priv)
 {
 	WacomCommonPtr common = priv->common;
-	WacomDeviceState ds[2] = {{0}};
+	WacomDeviceState ds[2] = {};
 	int midPoint_new = 0;
 	int midPoint_old = 0;
 	int i = 0, dist = 0;
 	WacomFilterState filterd;  /* borrow this struct */
 	int max_spread = common->wcmGestureParameters.wcmMaxScrollFingerSpread;
+	int gestureStart = 0;
 
 	if (!common->wcmGesture)
 		return;
@@ -640,6 +641,7 @@ static void wcmFingerScroll(WacomDevicePtr priv)
 				/* left button might be down. Send it up first */
 				wcmSendButtonClick(priv, 1, 0);
 				common->wcmGestureMode = GESTURE_SCROLL_MODE;
+				gestureStart = 1;
 			}
 		}
 	}
@@ -648,6 +650,13 @@ static void wcmFingerScroll(WacomDevicePtr priv)
 	if (common->wcmGestureMode != GESTURE_SCROLL_MODE)
 		return;
 
+	/* forget history leading up to the beginning of the gesture */
+	if (gestureStart)
+	{
+		common->wcmGestureState[0] = ds[0];
+		common->wcmGestureState[1] = ds[1];
+	}
+
 	/* initialize the points so we can rotate them */
 	filterd.x[0] = ds[0].x;
 	filterd.y[0] = ds[0].y;
@@ -711,11 +720,11 @@ static void wcmFingerScroll(WacomDevicePtr priv)
 static void wcmFingerZoom(WacomDevicePtr priv)
 {
 	WacomCommonPtr common = priv->common;
-	WacomDeviceState ds[2] = {{0}};
+	WacomDeviceState ds[2] = {};
 	int count, button;
-	int dist = touchDistance(common->wcmGestureState[0],
-			common->wcmGestureState[1]);
+	int dist;
 	int max_spread = common->wcmGestureParameters.wcmMaxScrollFingerSpread;
+	int gestureStart = 0;
 
 	if (!common->wcmGesture)
 		return;
@@ -738,13 +747,21 @@ static void wcmFingerZoom(WacomDevicePtr priv)
 			/* fingers moved apart more than 3 times
 			 * wcmMaxScrollFingerSpread, zoom mode is entered */
 			common->wcmGestureMode = GESTURE_ZOOM_MODE;
+			gestureStart = 1;
 		}
 	}
 
 	if (common->wcmGestureMode != GESTURE_ZOOM_MODE)
 		return;
 
-	dist = touchDistance(ds[0], ds[1]) - dist;
+	/* forget history leading up to the beginning of the gesture */
+	if (gestureStart)
+	{
+		common->wcmGestureState[0] = ds[0];
+		common->wcmGestureState[1] = ds[1];
+	}
+
+	dist = touchDistance(ds[0], ds[1]) - touchDistance(common->wcmGestureState[0], common->wcmGestureState[1]);
 	count = (int)((1.0 * abs(dist)/common->wcmGestureParameters.wcmZoomDistance) + 0.5);
 
 	/* user might have changed from left to right or vice versa */
