@@ -45,7 +45,7 @@ typedef struct {
 } wcmUSBData;
 
 static Bool usbDetect(InputInfoPtr);
-static Bool usbWcmInit(InputInfoPtr pDev, char* id, float *version);
+static Bool usbWcmInit(InputInfoPtr pDev, char* id, size_t id_len, float *version);
 static int usbProbeKeys(InputInfoPtr pInfo);
 static int usbStart(InputInfoPtr pInfo);
 static void usbInitProtocol5(WacomCommonPtr common, const char* id,
@@ -80,9 +80,9 @@ static struct _WacomModel mname =		\
 	.Start = usbStart,			\
 	.Parse = usbParse,			\
 	.DetectConfig = usbDetectConfig,	\
-};
+}
 
-DEFINE_MODEL(usbUnknown,	"Unknown USB",		5)
+DEFINE_MODEL(usbUnknown,	"Unknown USB",		5);
 DEFINE_MODEL(usbPenPartner,	"USB PenPartner",	4);
 DEFINE_MODEL(usbGraphire,	"USB Graphire",		4);
 DEFINE_MODEL(usbGraphire2,	"USB Graphire2",	4);
@@ -391,7 +391,7 @@ void usbListModels(void)
 }
 
 
-static Bool usbWcmInit(InputInfoPtr pInfo, char* id, float *version)
+static Bool usbWcmInit(InputInfoPtr pInfo, char* id, size_t id_len, float *version)
 {
 	int i;
 	struct input_id sID;
@@ -414,7 +414,7 @@ static Bool usbWcmInit(InputInfoPtr pInfo, char* id, float *version)
 
 	/* fetch vendor, product, and model name */
 	ioctl(pInfo->fd, EVIOCGID, &sID);
-	ioctl(pInfo->fd, EVIOCGNAME(sizeof(id)), id);
+	ioctl(pInfo->fd, EVIOCGNAME(id_len), id);
 
 	for (i = 0; i < ARRAY_SIZE(WacomModelDesc); i++)
 	{
@@ -546,6 +546,11 @@ int usbWcmGetRanges(InputInfoPtr pInfo)
 
 	if (!ISBITSET(ev,EV_ABS))
 	{
+		/* may be an expresskey only interface */
+		if (ISBITSET(common->wcmKeys, BTN_FORWARD) ||
+		    ISBITSET(common->wcmKeys, BTN_0))
+			return Success;
+
 		xf86Msg(X_ERROR, "%s: no abs bits.\n", pInfo->name);
 		return !Success;
 	}
@@ -560,6 +565,11 @@ int usbWcmGetRanges(InputInfoPtr pInfo)
 	/* max x */
 	if (ioctl(pInfo->fd, EVIOCGABS(ABS_X), &absinfo) < 0)
 	{
+		/* may be a PAD only interface */
+		if (ISBITSET(common->wcmKeys, BTN_FORWARD) ||
+		    ISBITSET(common->wcmKeys, BTN_0))
+			return Success;
+
 		xf86Msg(X_ERROR, "%s: unable to ioctl xmax value.\n", pInfo->name);
 		return !Success;
 	}
@@ -755,8 +765,7 @@ int usbWcmGetRanges(InputInfoPtr pInfo)
 
 	if (ioctl(pInfo->fd, EVIOCGBIT(EV_SW, sizeof(sw)), sw) < 0)
 	{
-		xf86Msg(X_ERROR, "%s: usbProbeKeys unable to ioctl "
-			"sw bits.\n", pInfo->name);
+		xf86Msg(X_ERROR, "%s: unable to ioctl sw bits.\n", pInfo->name);
 		return 0;
 	}
 	else if (ISBITSET(sw, SW_MUTE_DEVICE))
@@ -1117,7 +1126,7 @@ static int usbIdToType(int id)
  * Protocol 5 devices report different IDs for different styli and pucks,
  * Protocol 4 devices simply report STYLUS_DEVICE_ID, etc.
  *
- * @device_id id of the device
+ * @param device_id id of the device
  * @return device type
  */
 static int usbFindDeviceTypeById(int device_id)
@@ -1137,7 +1146,6 @@ static int usbFindDeviceTypeById(int device_id)
 		default: /* protocol 5 */
 			return usbIdToType(device_id);
 	}
-	return 0;
 }
 
 static void usbParseAbsEvent(WacomCommonPtr common,
@@ -1782,6 +1790,28 @@ static void usbDispatchEvents(InputInfoPtr pInfo)
 		return;
 	}
 
+	/* verify we have minimal data when entering prox */
+	if (ds->proximity && !dslast.proximity) {
+		struct input_absinfo absinfo;
+
+		if (!ds->x && !IsPad(priv)) {
+			if (ioctl(priv->pInfo->fd, EVIOCGABS(ABS_X), &absinfo) < 0)
+			{
+				DBG(-1, common, "unable to ioctl current x value.\n");
+				return;
+			}
+			ds->x = absinfo.value;
+		}
+		if (!ds->y && !IsPad(priv)) {
+			if (ioctl(priv->pInfo->fd, EVIOCGABS(ABS_Y), &absinfo) < 0)
+			{
+				DBG(-1, common, "unable to ioctl current x value.\n");
+				return;
+			}
+			ds->y = absinfo.value;
+		}
+	}
+
 	/*reset the serial number when the tool is going out */
 	if (!ds->proximity)
 		private->wcmLastToolSerial = 0;
