@@ -17,7 +17,6 @@
  */
 
 #ifdef HAVE_CONFIG_H
-#define WACOM_TOOLS
 #include "config.h"
 #endif
 
@@ -44,6 +43,9 @@
 #define TRACE(...) \
 	if (verbose) fprintf(stderr, "... " __VA_ARGS__)
 
+#define safe_realloc(p, n, s) \
+	((size_t)-1 / (n) < (s) ? NULL : realloc((p), (n)*(s)))
+
 static int verbose = False;
 
 enum printformat {
@@ -463,13 +465,13 @@ static param_t parameters[] =
 		.get_func = get_all,
 		.prop_flags = PROP_FLAG_READONLY,
 	},
-	{ NULL }
+	{}
 };
 
 /**
  * Deprecated parameters and their respective replacements.
  */
-struct deprecated
+static struct deprecated
 {
 	const char *name;
 	const char *replacement;
@@ -591,6 +593,8 @@ static struct modifier specialkeys[] = {
 	{"del", "Delete"}, {"home", "Home"},
 	{"end", "End"},
 
+	{"insert", "Insert"},
+
 	{"`", "quoteleft"},
 	{"-", "minus"},
 	{"=", "equal"},
@@ -655,6 +659,7 @@ static void print_value(param_t *param, const char *msg, ...)
 			vprintf(msg, va_args);
 			printf("\"\n");
 			break;
+		case FORMAT_DEFAULT:
 		default:
 			vprintf(msg, va_args);
 			printf("\n");
@@ -796,7 +801,7 @@ static void list_one_device(Display *dpy, XDeviceInfo *info)
 			if (atoms[j] == wacom_prop)
 				break;
 
-		if (j <= natoms)
+		if (j < natoms)
 		{
 			unsigned char	*data;
 			Atom		type;
@@ -946,7 +951,7 @@ static int special_map_modetoggle(Display *dpy, int argc, char **argv, unsigned
 static int special_map_displaytoggle(Display *dpy, int argc, char **argv, unsigned long *ndata, unsigned long *data, const size_t size);
 
 /* Valid keywords for the --set ButtonX options */
-struct keywords {
+static struct keywords {
 	const char *keyword;
 	int (*func)(Display*, int, char **, unsigned long*, unsigned long *, const size_t size);
 } keywords[] = {
@@ -985,16 +990,17 @@ static int special_map_modetoggle(Display *dpy, int argc, char **argv, unsigned
 	return 0;
 }
 
+/* the "displaytoggle" keyword isn't supported anymore, we just have this
+   here to tell people that. */
 static int special_map_displaytoggle(Display *dpy, int argc, char **argv, unsigned long *ndata, unsigned long *data, const size_t size)
 {
-	if (*ndata + 1 > size) {
-		fprintf(stderr, "Insufficient space to store all commands.\n");
-		return 0;
+	static int once_only = 1;
+	if (once_only)
+	{
+		printf ("Note: The \"displaytoggle\" keyword is not supported "
+			"anymore and will be ignored.\n");
+		once_only = 0;
 	}
-	data[*ndata] = AC_DISPLAYTOGGLE;
-
-	*ndata += 1;
-
 	return 0;
 }
 
@@ -1087,7 +1093,7 @@ static int keysym_to_keycode(Display *dpy, KeySym sym)
 		int i;
 
 		for (i = 0; i < XkbKeyGroupWidth(xkb, kc, state.group); i++)
-			if (XKeycodeToKeysym(dpy, kc, i) == sym)
+			if (XkbKeycodeToKeysym(dpy, kc, state.group, i) == sym)
 				goto out;
 	}
 
@@ -1153,10 +1159,12 @@ static int special_map_keystrokes(Display *dpy, int argc, char **argv, unsigned
 			key = convert_specialkey(key);
 			ks = XStringToKeysym(key);
 			if (ks == 0)
-				fprintf(stderr, "Warning: unable to map '%s' to a keycode.\n", key);
+				fprintf(stderr, "Warning: unable to map '%s' to a keysym.\n", key);
 		}
 
 		kc = keysym_to_keycode(dpy, ks);
+		if (kc == 0)
+			fprintf(stderr, "Warning: unable to map '%s' to a keycode.\n", key);
 
 		if (need_press && need_release && *ndata + nitems + 2 > size) {
 			fprintf(stderr, "Insufficient space to store all commands.\n");
@@ -1190,33 +1198,29 @@ static int special_map_keystrokes(Display *dpy, int argc, char **argv, unsigned
  */
 static char** strjoinsplit(int argc, char **argv, int *nwords)
 {
-	char buff[1024] = { 0 };
-	char **words	= NULL;
-	char *tmp, *tok;
-
-	while(argc--)
-	{
-		if (strlen(buff) + strlen(*argv) + 1 >= sizeof(buff))
-			break;
-
-		strcat(buff, *argv);
-		strcat(buff, " ");
-		argv++;
-	}
+	char **words = NULL;
+	int i, n;
 
 	*nwords = 0;
+	for (i = 0; i < argc; i++) {
+		char *tok = strtok(argv[i], " ");
+		while (tok) {
+			char **p = safe_realloc(words, *nwords+1, sizeof(char*));
+			if (!p) {
+				fprintf(stderr, "Unable to reallocate memory.\n");
+				return words;
+			}
 
-	for (tmp = buff; tmp && *tmp != '\0'; tmp = index((const char*)tmp, ' ') + 1)
-		(*nwords)++;
+			words = p;
+			words[*nwords] = strdup(tok);
+			if (!words[*nwords]) {
+				fprintf(stderr, "Unable to allocate memory.\n");
+				return words;
+			}
 
-	words = calloc(*nwords, sizeof(char*));
-
-	*nwords = 0;
-	tok = strtok(buff, " ");
-	while(tok)
-	{
-		words[(*nwords)++] = strdup(tok);
-		tok = strtok(NULL, " ");
+			(*nwords)++;
+			tok = strtok(NULL, " ");
+		}
 	}
 
 	return words;
@@ -1248,7 +1252,7 @@ static Bool parse_actions(Display *dpy, int argc, char **argv, unsigned long* da
 
 	if (nwords==1 && sscanf(words[0], "%d", &i) == 1)
 	{ /* Mangle "simple" button maps into proper actions */
-		char **new_words = realloc(words, 2);
+		char **new_words = realloc(words, sizeof(char*)*2);
 		if (new_words == NULL)
 		{
 			fprintf(stderr, "Unable to reallocate memory.\n");
@@ -1288,6 +1292,7 @@ static Bool parse_actions(Display *dpy, int argc, char **argv, unsigned long* da
 		if (!keyword_found)
 		{
 			fprintf(stderr, "Cannot parse keyword '%s' at position %d\n", words[i], i+1);
+			free(words);
 			return False;
 		}
 	}
@@ -1324,7 +1329,7 @@ static void special_map_property(Display *dpy, XDevice *dev, Atom btnact_prop, i
 	unsigned long btnact_nitems, bytes_after;
 	unsigned long nitems = 0;
 
-	data = calloc(256, sizeof(long));
+	data = calloc(256, sizeof(unsigned long));
 	if (!parse_actions(dpy, argc, argv, data, &nitems, 256))
 		goto out;
 
@@ -1753,14 +1758,15 @@ static void get_mode(Display *dpy, XDevice *dev, param_t* param, int argc, char
 	XValuatorInfoPtr v;
 
 	info = XListInputDevices(dpy, &ndevices);
-	while(ndevices--)
+
+	for (i = 0; i < ndevices; i++)
 	{
-		d = &info[ndevices];
+		d = &info[i];
 		if (d->id == dev->device_id)
 			break;
 	}
 
-	if (!ndevices) /* device id 0 is reserved and can't be our device */
+	if (i >= ndevices)
 	{
 		fprintf(stderr, "Unable to locate device.\n");
 		return;
@@ -1904,7 +1910,7 @@ static int get_actions(Display *dpy, XDevice *dev,
 				if (last_type != current_type)
 					strcat(buff, "key ");
 				is_press = !!(action & AC_KEYBTNPRESS);
-				detail = XKeycodeToKeysym(dpy, detail, 0);
+				detail = XkbKeycodeToKeysym(dpy, detail, 0, 0);
 				break;
 			case AC_BUTTON:
 				if (last_type != current_type)
@@ -2093,7 +2099,7 @@ static Bool need_xinerama(Display *dpy)
  * @param y_org[out]   Offset from the desktop origin to the mapped area's top edge
  * @return             True if the function could determine the mapped area
  */
-Bool get_mapped_area(Display *dpy, XDevice *dev, int *width, int *height, int *x_org, int *y_org)
+static Bool get_mapped_area(Display *dpy, XDevice *dev, int *width, int *height, int *x_org, int *y_org)
 {
 	Atom matrix_prop = XInternAtom(dpy, "Coordinate Transformation Matrix", True);
 	Atom type;
@@ -2117,7 +2123,7 @@ Bool get_mapped_area(Display *dpy, XDevice *dev, int *width, int *height, int *x
 	                   AnyPropertyType, &type, &format, &nitems,
 	                   &bytes_after, (unsigned char**)&data);
 
-	if (format != 32 || type != XInternAtom(dpy, "FLOAT", True))
+	if (format != 32 || type != XInternAtom(dpy, "FLOAT", True) || nitems != 9)
 	{
 		fprintf(stderr,"Property for '%s' has unexpected type - this is a bug.\n",
 			"Coordinate Transformation Matrix");
@@ -2265,14 +2271,14 @@ static Bool set_output_area(Display *dpy, XDevice *dev,
  */
 static Bool set_output_xrandr(Display *dpy, XDevice *dev, char *output_name)
 {
-	int i, found = 0;
+	int i, success = 0;
 	int x, y, width, height;
 	XRRScreenResources *res;
 	XRROutputInfo *output_info;
 	XRRCrtcInfo *crtc_info;
 
 	res = XRRGetScreenResources(dpy, DefaultRootWindow(dpy));
-	for (i = 0; i < res->noutput && !found; i++)
+	for (i = 0; i < res->noutput; i++)
 	{
 		output_info = XRRGetOutputInfo(dpy, res, res->outputs[i]);
 
@@ -2293,24 +2299,18 @@ static Bool set_output_xrandr(Display *dpy, XDevice *dev, char *output_name)
 
 		if (strcmp(output_info->name, output_name) == 0)
 		{
-			found = 1;
+			TRACE("Setting CRTC %s\n", output_name);
+			success = set_output_area(dpy, dev, x, y, width, height);
 			break;
 		}
 	}
 	XRRFreeScreenResources(res);
 
-	/* crtc holds our screen info, need to compare to actual screen size */
-	if (found)
-	{
-		TRACE("Setting CRTC %s\n", output_name);
-		return set_output_area(dpy, dev, x, y, width, height);
-	} else
-	{
+	if (!success)
 		printf("Unable to find output '%s'. "
 			"Output may not be connected.\n", output_name);
 
-		return False;
-	}
+	return success;
 }
 
 /**
